/*
 * la2vcd: convert logic analyzer waveforms to a VCD file
 *
 * Main program
 * $Id: la2vcd.c 14 2005-04-08 06:19:05Z eric $
 * Copyright 2005 Eric Smith <eric@brouhaha.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  Note that permission is
 * not granted to redistribute this program under the terms of any
 * other version of the General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */


#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


char *progname;


void usage (FILE *f)
{
  fprintf (f, "usage:\n");
  fprintf (f, "  %s [options] signalfile...]\n", progname);
  fprintf (f, "options:\n");
  fprintf (f, "  -ta <timefile>     times from file (mutually exclusive with -td)\n");
  fprintf (f, "  -td <time>         set delta time\n");
  fprintf (f, "  -r <radix>         set radix of next input file\n");
  fprintf (f, "  -b <count>         set bit count of next input file\n");
  fprintf (f, "  -s                 split next input file into individual bits\n");
  fprintf (f, "  -sf <bits> <name>  split a field from the next input file\n");
  fprintf (f, "  -ts <time>         set timescale of VCD file (default 1 ns)\n");
  fprintf (f, "  -mn <modulename>   set module name of signals in VCD file\n");
  fprintf (f, "  -o <vcdfile>       write VCD output to file (default stdout)\n");
  fprintf (f, "syntax:\n");
  fprintf (f, "  <radix> ::= 2 | 8 | 16\n");
  fprintf (f, "  <time> ::= <integer><unit>\n");
  fprintf (f, "  <unit> ::= s|ms|us|ns|ps|fs\n");
  fprintf (f, "  <bits> ::= <bitr>|<bitr>,<bits>\n");
  fprintf (f, "  <bitr> ::= <integer>|<integer>-<integer>\n");
}


void fatal (int ret, char *format, ...)
{
  va_list ap;

  fprintf (stderr, "fatal error: ");
  va_start (ap, format);
  vfprintf (stderr, format, ap);
  va_end (ap);
  if (ret == 1)
    usage (stderr);
  exit (ret);
}


void *alloc (size_t size)
{
  void *p;

  p = calloc (1, size);
  if (! p)
    fatal (2, "Memory allocation failed\n");
  return (p);
}


char *newstr (char *orig)
{
  int len;
  char *r;

  len = strlen (orig);
  r = (char *) alloc (len + 1);
  memcpy (r, orig, len + 1);
  return (r);
}


void strip_whitespace (char *p)
{
  int len = strlen (p);
  char *p2 = p;

  while (len && isspace (p [len - 1]))
    p [--len] = '\0';

  while ((*p2) && isspace (*p2))
    {
      p2++;
      len--;
    }

  if (p != p2)
    memmove (p, p2, len);
}


static int bits_set (uint64_t x)
{
  int count = 0;

  while (x)
    {
      x &= ~ - x;
      count++;
    }

  return count;
}


static int significant_bits (uint64_t x)
{
  int count = 0;
  uint64_t i = 0;

  while (i < x)
    {
      i = (i << 1) + 1;
      count++;
    }
  return count;
}


static uint64_t extract_and_pack_bits (uint64_t src, uint64_t mask)
{
  uint64_t dest = 0;
  uint64_t db = 1;
  uint64_t sb = 1;

  while (mask)
    {
      if (mask & sb)
	{
	  if (src & sb)
	    {
	      dest |= db;
	    }
	  mask &= ~ sb;
	  db <<= 1;
	}
      sb <<= 1;
    }
  return dest;
}


static void fprint_binary (FILE *o, uint64_t val, int bits)
{
  int bit;
  bool leading = true;
  while (--bits >= 0)
    {
      bit = ((val & (1LL << bits)) != 0LL);
      if (leading && (bits != 0) && ! bit)
	continue;
      leading = false;
      fputc ('0' + bit, o);
    }
}


FILE *timef;
FILE *outf;

char *module_name;


typedef struct signal_info_t
{
  struct signal_info_t *next;

  struct input_file_info_t *input_file;
  uint64_t source_mask;  // which bits from input file are extracted and
                         // packed to get the value of this signal
  int bit_count;         // total bits in signal

  int max_index;         // index range when displayed as vector
  int min_index;

  char *name;
  char *vcd_sym;

  uint64_t value;
  uint64_t prev_value;
} signal_info_t;


signal_info_t *first_signal;
signal_info_t *last_signal;


typedef struct input_file_info_t
{
  struct input_file_info_t *next;
  FILE *inf;
  int radix;
  int bit_count;
  char *name;
  char *buf;
  uint64_t allocated_mask;  // bits which are allocated to signals
  uint64_t value;
} input_file_info_t;


input_file_info_t *first_input_file;
input_file_info_t *last_input_file;


#define INPUT_FILE_BUF_SIZE 255


double time_delta = 0;
double time_scale = 0;
char time_scale_string [7];

double time_offset = 0;

double current_time;
double prev_time;


static signal_info_t *alloc_new_signal (void)
{
  signal_info_t *signal = alloc (sizeof (signal_info_t));
  if (first_signal)
    last_signal->next = signal;
  else
    first_signal = signal;
  last_signal = signal;
  return signal;
}



#define MAX_VCD_SYM_LEN 3

static char* new_vcd_sym (void)
{
  int i;
  static int chars = 1;
  static char vcd_sym [MAX_VCD_SYM_LEN + 1] = { ' ', 0 };

  for (i = chars - 1; i >= 0; i--)
    {
      if ((++vcd_sym [i]) <= '~')
	break;
      vcd_sym [i] = '!';
    }

  if (i < 0)
    {
      if ((++chars) > MAX_VCD_SYM_LEN)
	fatal (4, "Too many signals for VCD file\n");
      memset (vcd_sym, '!', chars - 1);
    }

  return newstr (vcd_sym);
}


static void write_vcd_header (void)
{
  time_t t;
  signal_info_t *signal;
  char buf [100];

  t = time (NULL);
  strftime (buf, sizeof (buf), "%b %d, %Y  %H:%M:%S", localtime (& t));

  fprintf (outf, "$date\n");
  fprintf (outf, "    %s\n", buf);
  fprintf (outf, "$end\n");
  fprintf (outf, "$version\n");
  fprintf (outf, "    TOOL: la2vcd 0.02\n");
  fprintf (outf, "$end\n");
  fprintf (outf, "$time_scale\n");
  fprintf (outf, "    %s\n", time_scale_string);
  fprintf (outf, "$end\n");
  fprintf (outf, "$scope module %s $end\n",
	   module_name ? module_name : "logic_analyzer");
  for (signal = first_signal; signal; signal = signal->next)
    {
      signal->vcd_sym = new_vcd_sym ();
      if (signal->bit_count == 1)
	fprintf (outf, "$var wire 1 %s %s [%d] $end\n",
		 signal->vcd_sym,
		 signal->name,
		 signal->min_index);
      else
	fprintf (outf, "$var wire %d %s %s [%d:%d] $end\n",
		 signal->bit_count, 
		 signal->vcd_sym,
		 signal->name,
		 signal->max_index,
		 signal->min_index);
    }
  fprintf (outf, "$upscope $end\n");
  fprintf (outf, "$enddefinitions $end\n");
}


static void write_time (void)
{
  fprintf (outf, "#%lld\n", (int64_t) (current_time / time_scale));
}


static void write_vcd_data (void)
{
  static bool first_time = true;
  bool wrote_time = false;
  signal_info_t *signal;

  if (first_time)
    {
      write_time ();
      fprintf (outf, "$dumpvars\n");
    }

  for (signal = first_signal; signal; signal = signal -> next)
    {
      signal->prev_value = signal->value;
      signal->value = extract_and_pack_bits (signal->input_file->value,
					     signal->source_mask);
      if (first_time || (signal->value != signal->prev_value))
	{
	  if ((! first_time) && (! wrote_time))
	    {
	      write_time ();
	      wrote_time = true;
	    }
	  if (signal->bit_count == 1)
	    fprintf (outf, "%lld", signal->value & 1);
	  else
	    {
	      fprintf (outf, "b");
	      fprint_binary (outf, signal->value, signal->bit_count);
	      fprintf (outf, " ");
	    }
	  fprintf (outf, "%s\n", signal->vcd_sym);
	}
    }

  if (first_time)
    {
      fprintf (outf, "$end\n");
      first_time = false;
    }
}


static void write_vcd_trailer (void)
{
  signal_info_t *signal;

  fprintf (outf, "$dumpoff\n");
  for (signal = first_signal; signal; signal = signal -> next)
    if (signal->bit_count == 1)
      fprintf (outf, " x%s\n", signal->vcd_sym);
    else
      fprintf (outf, "bx %s\n", signal->vcd_sym);
  fprintf (outf, "$end\n");
}


static void read_timestamp_file_header (void)
{
  if (timef)
    {
      char ts_buf [100];
      fgets (ts_buf, sizeof (ts_buf), timef);
      if (ferror (timef))
	fatal (4, "error reading timestamp file\n");
      if (feof (timef))
	fatal (4, "eof reading timestamp file\n");
    }
}


static void read_input_file_header (input_file_info_t *input_file)
{
  fgets (input_file->buf, INPUT_FILE_BUF_SIZE, input_file->inf);
      
  if (ferror (input_file->inf))
    fatal (4, "error reading input file headers\n");

  if (feof (input_file->inf))
    fatal (4, "eof while reading input file headers\n");

  strip_whitespace (input_file->buf);

  input_file->name = newstr (input_file->buf);
}


// returns true if OK (not EOF)
static bool advance_time (void)
{
  bool eof = false;
  prev_time = current_time;
  if (timef)
    {
      char ts_buf [100];
      fgets (ts_buf, sizeof (ts_buf), timef);
      if (ferror (timef))
	fatal (4, "error reading timestamp file\n");
      eof = feof (timef);
      if (! eof)
	current_time = atof (ts_buf) - time_offset;
    }
  else
    current_time = prev_time + time_delta;
  return ! eof;
}


// returns true if OK (not EOF)
static bool read_input_file_sample (input_file_info_t *input_file)
{
  fgets (input_file->buf, INPUT_FILE_BUF_SIZE, input_file->inf);
      
  if (ferror (input_file->inf))
    fatal (4, "error reading input file '%s'\n", input_file->name);

  if (feof (input_file->inf))
    return false;

  strip_whitespace (input_file->buf);

  input_file->value = strtoull (input_file->buf, NULL, input_file->radix);
  input_file->value &= ((1LL << input_file->bit_count) - 1);

  return true;
}


static bool read_one_line (void)
{
  input_file_info_t *input_file;
  int eof_count = 0;
  int non_eof_count = 0;

  for (input_file = first_input_file; input_file; input_file = input_file -> next)
    if (read_input_file_sample (input_file))
      non_eof_count++;
    else
      eof_count++;

  if (eof_count && ! non_eof_count)
    return false;  // done

  if (eof_count)
    fatal (3, "files contain differing line counts\n");

  return true;  // not yet done
}


void close_all_files (void)
{
  input_file_info_t *input_file;

  fclose (outf);
  if (timef)
    fclose (timef);
  for (input_file = first_input_file; input_file; input_file = input_file -> next)
    fclose (input_file->inf);
}


static signal_info_t *split_signals_from_input (input_file_info_t *input_file)
{
  signal_info_t *first_new_signal = NULL;
  int bit;
  for (bit = input_file->bit_count - 1; bit >= 0; bit--)
    {
      signal_info_t *signal = alloc_new_signal ();
      signal->bit_count = 1;
      signal->max_index = bit;
      signal->min_index = bit;
      signal->source_mask = 1ULL << bit;
      signal->name = input_file->name;
      if (! first_new_signal)
	first_new_signal = signal;
    }
  return first_new_signal;
}


static signal_info_t *split_field (char *bits, char *name)
{
  signal_info_t *signal;
  unsigned long i, j;

  signal = alloc_new_signal ();
  signal->name = name;

  signal->source_mask = 0;

  while (*bits)
    {
      i = strtoul (bits, & bits, 10);
      if ((i < 0) || (i > 63))
	fatal (1, "bit out of range\n");
      switch (*bits)
	{
	case '-':
	  j = strtoul (bits + 1, & bits, 10);
	  if ((j < 0) || (j > 63))
	    fatal (1, "bit out of range\n");
	  if (j > i)
	    fatal (1, "bit ranges must be from larger index to smaller index\n");
	  signal->source_mask |= (((1ULL << i) - 1ULL) & ~ ((1ULL << j) - 1ULL));
	  break;
	case ',':
	  bits++;
	  // FALL INTO NEXT CASE
	case '\0':
	  signal->source_mask |= (1ULL << i);
	  break;
	default:
	  fatal (1, "bad character '%c' in bit definition\n", *bits);
	}
    }

  signal->bit_count = bits_set (signal->source_mask);
  signal->max_index = signal->bit_count - 1;
  signal->min_index = 0;

  return (signal);
}


static double parse_time (char *p, double *numeric, char **unit)
{
  double t;
  double sc;
  char *p2;

  t = strtod (p, & p2);
  if (t == 0)
    t = 1.0;

  if (strcmp (p2, "fs") == 0)
    sc = 1.0e-15;
  else if (strcmp (p2, "ps") == 0)
    sc = 1.0e-12;
  else if (strcmp (p2, "ns") == 0)
    sc = 1.0e-9;
  else if (strcmp (p2, "us") == 0)
    sc = 1.0e-6;
  else if (strcmp (p2, "ms") == 0)
    sc = 1.0e-3;
  else if (strcmp (p2, "s") != 0)
    fatal (1, "invalid time specification\n");

  if (numeric)
    *numeric = t;
  if (unit)
    *unit = p2;

  return (t * sc);
}


static double parse_time_scale (char *p, char *str)
{
  double result;
  double numeric;
  char *unit;

  result = parse_time (p, & numeric, & unit);
  if ((numeric != 1.0) && (numeric != 10.0) && (numeric != 100.0))
    fatal (1, "time scale numeric part must be 1, 10, or 100\n");
  sprintf (str, "%d %s", (int) numeric, unit);
  return (result);
} 


int main (int argc, char *argv [])
{
  progname = argv [0];
  outf = stdout;
  int next_radix = 0;
  int next_bit_count = 0;
  bool next_split = false;
  signal_info_t *new_signals = NULL;

  while (--argc)
    {
      argv++;
      if (argv [0][0] == '-')
	{
	  if ((strcmp (argv [0], "-o") == 0) ||
	      (strcmp (argv [0], "--output-file") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-o option must be followed by output filename\n");
	      if (outf != stdout)
		fatal (1, "only one output file may be specified\n");
	      outf = fopen (argv [1], "w");
	      if (! outf)
		fatal (2, "can't open output file\n");
	      argc--;
	      argv++;
	    }
	  else if ((strcmp (argv [0], "-ta") == 0) ||
		   (strcmp (argv [0], "--time-file-abs") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-ta option must be followed by timestamp filename\n");
	      if (timef)
		fatal (1, "only one timestamp file may be specified\n");
	      if (time_delta)
		fatal (1, "-ta and -td options are mutually exclusive\n");
	      timef = fopen (argv [1], "r");
	      if (! timef)
		fatal (2, "can't open timestamp file\n");
	      read_timestamp_file_header ();
	      argc--;
	      argv++;
	    }
	  else if ((strcmp (argv [0], "-td") == 0) ||
		   (strcmp (argv [0], "--time-delta") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-td option must be followed by time\n");
	      if (time_delta)
		fatal (1, "only one time delta may be specified\n");
	      if (timef)
		fatal (1, "-ta and -td options are mutually exclusive\n");
	      time_delta = parse_time (argv [1], NULL, NULL);
	      argc--;
	      argv++;
	    }
	  else if ((strcmp (argv [0], "-ts") == 0) ||
		   (strcmp (argv [0], "--time-scale") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-ts option must be followed by time scale\n");
	      if (time_scale)
		fatal (1, "only one time scale may be specified\n");
	      time_scale = parse_time_scale (argv [1], time_scale_string);
	      argc--;
	      argv++;
	    }
	  else if ((strcmp (argv [0], "-r") == 0) ||
		   (strcmp (argv [0], "--radix") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-r option must be followed by radix\n");
	      next_radix = atoi (argv [1]);
	      if ((next_radix != 2) && (next_radix != 8) && (next_radix != 16))
		fatal (1, "invalid radix\n");
	      argc--;
	      argv++;
	    }
	  else if ((strcmp (argv [0], "-b") == 0) ||
		   (strcmp (argv [0], "--bit-count") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-b option must be followed by bit count\n");
	      next_bit_count = atoi (argv [1]);
	      if ((next_bit_count < 1) || (next_bit_count > 64))
		fatal (1, "bit count must be in range 1 through 64 inclusive\n");
	      argc--;
	      argv++;
	    }
	  else if ((strcmp (argv [0], "-s") == 0) ||
		   (strcmp (argv [0], "--split") == 0))
	    next_split = true;
	  else if ((strcmp (argv [0], "-sf") == 0) ||
		   (strcmp (argv [0], "--split-field") == 0))
	    {
	      signal_info_t *signal;
	      if (argc <= 2)
		fatal (1, "-sf option must be followed by bit(s) and name \n");
	      signal = split_field (argv [1], argv [2]);
	      if (! new_signals)
		new_signals = signal;
	      argc -= 2;
	      argv += 2;
	    }
	  else if ((strcmp (argv [0], "-mn") == 0) ||
		   (strcmp (argv [0], "--module-name") == 0))
	    {
	      if (argc <= 1)
		fatal (1, "-mn option must be followed by module name\n");
	      module_name = argv [1];
	      argc--;
	      argv++;
	    }
	  else
	    fatal (1, "unrecognized option '%s'\n", argv [0]);
	}
      else
	{
	  input_file_info_t *new_input_file = alloc (sizeof (input_file_info_t));
	  new_input_file->inf = fopen (argv [0], "r");
	  if (! new_input_file->inf)
	    fatal (2, "can't open input_file file '%s'\n", argv [0]);
	  new_input_file->radix = next_radix ? next_radix : 2;
	  new_input_file->bit_count = next_bit_count;
	  next_radix = 0;
	  next_bit_count = 0;
	  new_input_file->buf = alloc (INPUT_FILE_BUF_SIZE);
	  if (first_input_file)
	    last_input_file->next = new_input_file;
	  else
	    first_input_file = new_input_file;
	  last_input_file = new_input_file;

	  read_input_file_header (new_input_file);
	  if (! read_input_file_sample (new_input_file))
	    fatal (3, "no data in file\n");

	  if (! new_input_file->bit_count)
	    new_input_file->bit_count = strlen (new_input_file->buf) * significant_bits (new_input_file->radix - 1);

	  if (next_split)
	    {
	      new_signals = split_signals_from_input (new_input_file);
	      next_split = false;
	    }

	  if (! new_signals)
	    {
	      new_signals = alloc_new_signal ();
	      new_signals->bit_count = new_input_file->bit_count;
	      new_signals->max_index = new_input_file->bit_count - 1;
	      new_signals->min_index = 0;
	      new_signals->source_mask = (1ULL << new_signals->bit_count) - 1;
	      new_signals->name = new_input_file->name;
	    }

	  while (new_signals)
	    {
	      new_signals->input_file = new_input_file;
	      if (new_signals->source_mask & ~ ((1 << new_input_file->bit_count) - 1))
		fatal (3, "field bit positions out of range\n");
	      if (new_signals->source_mask & new_input_file->allocated_mask)
		fprintf (stderr, "warning: bits of input file allocated to multiple signals\n");
	      new_input_file->allocated_mask |= new_signals->source_mask;
	      new_signals = new_signals->next;
	    }
	}
    }

  if (! time_scale)
    time_scale = 1.0e-9;

  if (! first_input_file)
    fatal (1, "at least one input file file must be specified\n");

  if ((! timef) && (! time_delta))
    fatal (1, "either -ta or -td option must be given\n");

  write_vcd_header ();

  if (timef)
    {
      if (! advance_time ())  // get the first timestamp
	fatal (3, "no data in timestamp file\n");
      time_offset = current_time;  // first time becomes zero in the output
      current_time = 0;            // VCF file, so we use it as time_offset
    }

  write_vcd_data ();  // we've already got the first set of samples

  while (1)
    {
      advance_time ();
      if (read_one_line ())
	write_vcd_data ();
      else
	break;
    }

  write_vcd_trailer ();

  close_all_files ();

  exit (0);
}
